/*
          Copyright (C) 1993, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/*
** Modified for the RSNA'93 DICOM Demonstration by:
**	FB10 - Informatik
**	Oldenburg University
**	Germany
**
*/
/* 			snq.c
**
** This file contains routines that implement operations
** on the Study Name Queue.  The SNQ is saved in shared memory
** and the SNQ_Dequeue and SNQ_Enqueue procedures are used
** to access its elements. 
**
** SNQ_code SNQ_KillQueue ( int qid ) 
**	Frees up the shared memory and semaphore associated with the
**	named queue and removes the file descriptor.
** SNQ_code SNQ_InitQueue (int size)
**	This routine allocates shared memory for the SNQ data structure
**	and the two pointers to that data structure (SNQ_head and SNQ_tail).
**	It also initializes the semaphore to control friendly access to
**	the queue between multiple processes.  It then initializes the
**	queue so that it will accept data.
** SNQ_code SNQ_Enqueue (QUEUE_Element element)
**	This routine enqueues a queue element to the SNQ structure.
**	It waits for access to the queue by waiting for a 
**	semaphore to be released.  It checks that the
**	queue is not full and if it isn't adds a new element at
**	the position pointed to by SNQ_tail. It increments the
**	tail pointer and checks for wrap around.  Finally, the
**	routine uses a semaphore to signal that this process
**	is finished accessing the data structure. 
** SNQ_code SNQ_Dequeue (QUEUE_Element *element)
**	This routine dequeues a queue element from the SNQ structure.
**	It waits for access to the queue by waiting for a 
**	semaphore to be released.  It checks that the
**	queue is not empty, and if it isn't copys the element
**	at the position pointed to by SNQ_head.  It increments
**	the SNQ_head pointer and checks for wrap around. 
** SNQ_code SNQ_GetQueue ( void )
**	This routine obtains access to the currently defined queue
**	by accessing the file with the shared memory and semaphore
**	identifiers, and attaching these structures.
** SNQ_code SNQ_PrintQueue ()
**	This routines prints out the contents of the SNQ data
**	structure to stdout and prints the positions of the
**	SNQ_head and SNQ_tail pointers.
** SNQ_code SNQ_Wait (int sem_id)
**	This routines suspends execution of the calling process
**	until it receives a signal (from SNQ_Signal, for example). 
**	This routine is used to indicate that the calling process
**	is waiting for the semaphore to be released.
** SNQ_code SNQ_Signal (int sem_id)
** 	This routines sets the value of the semaphore associated
**	with SEM_ID to 1.  This routine is used to signal 
**	that the semaphore has been released. 
** char *SNQ_MakeFilename (int qid)
**	This routine makes a valid filename from the queue id and the
**	and the environment variable QUEUE_DIRECTORY.
**
**	$Log: snq.cc,v $
**	Revision 1.4  1996/06/17 12:56:03  meichel
**	Corrected text in some error messages.
**
**	Revision 1.3  1996/05/06 07:30:18  hewett
**	Changed some DEBUG output to only be generated in DEBUG2 is defined.
**	The output was messing up the ti program.
**
**	Revision 1.2  1996/04/18 15:29:50  meichel
**	Version for CAR'96
**
**	Revision 1.1  1996/03/28 17:08:03  meichel
**	first commit with new file structure
**
**	Revision 2.1  1995/03/17 13:52:50  meichel
**	Support for LINUX added. Compiles with -DLINUX.
**
 * Revision 2.0  1994/11/17  09:45:32  meichel
 * Version 2.0 for RSNA 1994 demonstration.
 * New features: Metaheader support, new SOP classes, lots of bugs fixed.
 *
 * Revision 1.6  1994/03/17  18:15:33  hewett
 * Changes due to must recent DICOM data dictionary (dcmcode)
 *
 * Revision 1.5  1994/03/17  09:28:30  hewett
 * Updated copyright notice.
 *
 * Revision 1.4  1994/03/16  16:03:31  hewett
 * Fixed dereferencing bug.
 *
 * Revision 1.3  1993/11/19  14:39:09  hewett
 * Fixed unitialized variable bug in SNQ-Wait which was causing lots
 * of Semaphore Failures.
 *
 * Revision 1.2  1993/10/14  09:42:08  hewett
 * Updated original MIR header for 1993
 *
**
**	$Id: snq.cc,v 1.4 1996/06/17 12:56:03 meichel Exp $
**
*/

#include "osconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/* first we check if the required semaphore and shared memory functions
   are available */

#if defined(HAVE_SEMCTL) && defined(HAVE_SEMGET) && defined (HAVE_SEMOP) && \
    defined(HAVE_SHMAT) && defined(HAVE_SHMDT) && defined(HAVE_SHMCTL) && \
    defined(HAVE_SHMGET) && defined(HAVE_SYS_TYPES_H)
#define IPC_AVAILABLE
#else
#undef IPC_AVAILABLE
#endif

#ifdef IPC_AVAILABLE

extern "C" 
{

#include <sys/types.h> /* we checked this before. Needed for key_t, size_t. */

#ifndef HAVE_DECLARATION_UNION_SEMUN
union semun 
{
  int val;
  void *buf; /* should be a struct semid_ds *, but we don't need it anyway. */
  ushort *array;
};
#endif

#ifndef HAVE_DECLARATION_STRUCT_SEMBUF
struct sembuf {
  ushort_t sem_num; /* semaphore # */
  short    sem_op;  /* semaphore operation */
  short    sem_flg; /* operation flags */
};
#endif

#ifdef HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif

#ifdef HAVE_SYS_SEM_H
#include <sys/sem.h>
#endif

#ifndef HAVE_PROTOTYPE_SEMCTL
     int semctl(int semid, int semnum, int cmd, ...);
#endif
#ifndef HAVE_PROTOTYPE_SEMGET
     int semget(key_t key, int nsems, int semflg);
#endif
#ifndef HAVE_PROTOTYPE_SEMOP
     int semop(int semid, struct sembuf *sops, size_t nsops);
#endif

#ifdef HAVE_SYS_SHM_H
#include <sys/shm.h>
#else
/* 
   This is a dummy declaration of struct shmid_ds. The real declaration from 
   sys/shm.h is very complex. Since we only pass "opaque" pointers of this 
   type to shmctl() and don't need the results, this hack should work.
   Anyway: Which weard kind of system
   should support shmctl() and have no appropriate header file??? 
 */
     struct shmid_ds { int dummy_element; };
#endif

#ifndef HAVE_PROTOTYPE_SHMAT
     void *shmat(int shmid, void *shmaddr, int shmflg);
#endif
#ifndef HAVE_PROTOTYPE_SHMDT
     int shmdt(void *shmaddr);
#endif
#ifndef HAVE_PROTOTYPE_SHMCTL
     int shmctl(int shmid, int cmd, struct shmid_ds *buf);
#endif
#ifndef HAVE_PROTOTYPE_SHMGET
     int shmget(key_t key, int size, int shmflg);
#endif
}

#endif /* IPC_AVAILABLE */

#include 	"snq.h"	

#ifdef IPC_AVAILABLE
/*
 * static structures for the queue...
 */
static QUEUE_Element  *SNQ;	/* Pointer to the queue elements	*/
static QUEUE_Pointers *QPTRS;	/* Pointer to the head, tail, and size	*/
static int            SHM_ID;	/* the shared memory id			*/
static int            SEM_ID;	/* the semaphore id			*/
static char           *SHM_ADD;/* the address of the shared memory	*/
/*
 * A flag to remember if we did a GetQueue yet...
 */
static int SM_attached = 0;
static union semun sem_buf;
#endif

#define SNQ_UNIMPLEMENTED_TXT	\
	"SNQ Routine %s is unimplemented on this machine architecture\n"

/* SNQ_KillQueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_KillQueue ( int qid ) {

#ifdef IPC_AVAILABLE

	struct	shmid_ds	*shm_buf = NULL;	
	char			command[512];

	if ((shmdt (SHM_ADD)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_KillQueue: shmdt failed\n");
#endif
		return (SNQ_SHAREDMEMORYFAIL);
	}
	SM_attached = 0;

	if ((shmctl (SHM_ID, IPC_RMID, shm_buf)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_KillQueue: shmctl failed\n");
#endif
		return (SNQ_SHAREDMEMORYFAIL);
	}
	if ((semctl (SEM_ID, 0, IPC_RMID, sem_buf)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_SemaphoreRelease: semctl failed\n");
#endif
		return (SNQ_SEMAPHOREFAIL);
	}

	strcpy(command,"rm -r ");
	strcat(command,SNQ_MakeFilename(qid));
	system (command);

	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_KillQueue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_InitQueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_InitQueue (int qid, int size){

#ifdef IPC_AVAILABLE

	int	q_size;		/* size of the queue and pointers	*/

	key_t	key;		/* key for the shared mem		*/

	char	*shmadd,	/* temp shared mem address		*/
		*fname;		/* possibly long filename for queue	*/

	FILE	*fd;

	struct shmid_ds *shm_buf = NULL;	/* In case we need to abort... */
/*
 * Acquire and attach to the shared memory...
 */
	key = 0;

#ifdef DEBUG2
	fprintf(stderr, "SNQ_InitQueue: in snq.o started\n");
#endif

	q_size = (size *sizeof (QUEUE_Element)) + sizeof (QUEUE_Pointers);
	if ((SHM_ID = shmget (key, (int)q_size, IPC_CREAT|0777)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: shmget failed\n");
#endif	
		return (SNQ_SHAREDMEMORYFAIL);
	}

	if(SM_attached){
		shmdt( SHM_ADD );
		SM_attached = 0;
	}

	SHM_ADD = 0;

	if ((shmadd = (char *)shmat (SHM_ID, SHM_ADD, SHM_RND)) == (char *) -1 ){
		shmctl (SHM_ID, IPC_RMID, shm_buf);
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: shmat failed\n");
#endif	
		return (SNQ_SHAREDMEMORYFAIL);
	}

/*
 * Save for later detaching if needed
 */
	SHM_ADD = shmadd;

/*
 * Acquire and initialize the semaphore...
 */
	if ((SEM_ID = semget (key, 1, IPC_CREAT|0777)) == -1){
		shmdt (SHM_ADD);
		shmctl (SHM_ID, IPC_RMID, shm_buf);
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: semget failed\n");
#endif	
		return (SNQ_SEMAPHOREFAIL);
	}
	sem_buf.val = 1;
	if ((semctl (SEM_ID, 0, SETVAL, sem_buf)) == 1){
		shmdt (SHM_ADD);
		shmctl (SHM_ID, IPC_RMID, shm_buf);
		semctl (SEM_ID, 0, IPC_RMID, sem_buf);
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: semctl failed\n");
#endif	
		return (SNQ_SEMAPHOREFAIL);
	}

/*
 * Map structure for head/tail pointers into shared memory 
 */
	QPTRS = (QUEUE_Pointers *)shmadd;
	shmadd += sizeof (QUEUE_Pointers);

/*
 * Map structure for queue into shared memory 
 */
	SNQ = (QUEUE_Element *)shmadd;

	QPTRS->SNQ_head = QPTRS->SNQ_tail = 0;
	QPTRS->SNQ_size = size;
/*
 * Create the file for passing the shared memory and semaphore ids
 */
	
	if ((fname = SNQ_MakeFilename (qid)) == NULL){
		shmdt (SHM_ADD);
		shmctl (SHM_ID, IPC_RMID, shm_buf);
		semctl (SEM_ID, 0, IPC_RMID, sem_buf);
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: Cannot access QUEUE_DIRECTORY/snq???.dat\n");
#endif	
		return (SNQ_FILEACCESSFAIL);
	}
        if ((fd = fopen (fname, "w")) == NULL){
		shmdt (SHM_ADD);
		shmctl (SHM_ID, IPC_RMID, shm_buf);
		semctl (SEM_ID, 0, IPC_RMID, sem_buf);
#ifdef DEBUG
		fprintf(stderr,"SNQ_InitQueue: Cannot access file '%s'\n",fname);
		perror("Error code");
#endif	
		return (SNQ_FILEACCESSFAIL);
	}
	fprintf (fd, "%d\n", SEM_ID);
	fprintf (fd, "%d\n", SHM_ID);
	fclose (fd);

	SM_attached = 1;
	return (SNQ_GetQueue( qid ));
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_InitQueue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_Enqueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_Enqueue (QUEUE_Element *element){


#ifdef IPC_AVAILABLE

	SNQ_code ret;


	if ( (ret = SNQ_Wait ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_Enqueue: Semaphore failure\n");
#endif	
		return (ret);
	}
	if ( (QPTRS->SNQ_head == (QPTRS->SNQ_tail+1)) || 
	    ((QPTRS->SNQ_head ==  0) && 
	     (QPTRS->SNQ_tail == (QPTRS->SNQ_size-1))) ){
		if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
			fprintf(stderr,"SNQ_Enqueue: Semaphore failure\n");
#endif	
			return (ret);
		}
		return (SNQ_QUEUEFULL);
	}
	SNQ [QPTRS->SNQ_tail] = *element;

	if (++QPTRS->SNQ_tail >= QPTRS->SNQ_size)
		QPTRS->SNQ_tail = 0;

	if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_Enqueue: Semaphore failure\n");
#endif	
		return (ret);
	}

#ifdef DEBUG2
	fprintf(stderr, "SNQ_Enqueue in snq.o: Association ID : %d\n",
		element->Q_AssociationID);
#endif
		return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_Enqueue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_Dequeue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_Dequeue (QUEUE_Element *element){

#ifdef IPC_AVAILABLE

	SNQ_code ret;
	if ( (ret = SNQ_Wait ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_Dequeue: Semaphore failure\n");
#endif	
		return (ret);
	}
	if (QPTRS->SNQ_head == QPTRS->SNQ_tail){
		if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
			fprintf(stderr,"SNQ_Dequeue: Semaphore failure\n");
#endif	
			return (ret);
		}
		return (SNQ_QUEUEEMPTY);
	}

/*
 * Copy the next item out of the queue
 */
	*element = SNQ [QPTRS->SNQ_head]; 

/*
 * Check for wrap-around unsigned long 
 */
	if (++QPTRS->SNQ_head >= QPTRS->SNQ_size)
   		QPTRS->SNQ_head = 0;

	if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_Dequeue: Semaphore failure\n");
#endif	
		return (ret);
	}


#ifdef DEBUG2
	fprintf(stderr, "SNQ_Dequeue in snq.o: Association ID : %d\n",
		element->Q_AssociationID);
#endif

	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_Dequeue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_GetQueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
** deb		31-Jul-1992		Do detach from previously attached
**					to segment if necessary.
*/

SNQ_code SNQ_GetQueue ( int qid ) {

#ifdef IPC_AVAILABLE

	FILE	*fd;
	char	*shmadd,
		*fname;

	if ((fname = SNQ_MakeFilename (qid)) == NULL){
#ifdef DEBUG
		fprintf(stderr,"SNQ_GetQueue: Cannot access QUEUE_DIRECTORY/snq???.dat\n");
#endif	
		return (SNQ_FILEACCESSFAIL);
	}
	if ((fd = fopen (fname, "r")) == NULL){
#ifdef DEBUG
		fprintf(stderr,"SNQ_GetQueue: Cannot access file '%s'\n",fname);
		perror("Error code");
#endif	
		return (SNQ_FILEACCESSFAIL);
 	}
	fscanf (fd, "%d", &SEM_ID);
        fscanf (fd, "%d", &SHM_ID);
	fclose (fd);

	if(SM_attached){
		shmdt( SHM_ADD );
		SM_attached = 0;
	}

	SHM_ADD = 0;
	if ((shmadd = (char *)shmat (SHM_ID, SHM_ADD, SHM_RND)) == (char *) -1 ){
#ifdef DEBUG
		fprintf(stderr,"SNQ_GetQueue: shmat failed\n");
#endif	
		return (SNQ_SHAREDMEMORYFAIL);
	}
	SHM_ADD = shmadd;
/*
 * Map structure for head/tail pointers into shared memory 
 */
	QPTRS = (QUEUE_Pointers *)shmadd;
	shmadd += sizeof (QUEUE_Pointers);

/*
 * Map structure for queue into shared memory 
 */
	SNQ = (QUEUE_Element *)shmadd;

	SM_attached = 1;
	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_GetQueue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_DumpQueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_DumpQueue (QUEUE_Element **table, int *num_active){

#ifdef IPC_AVAILABLE

	SNQ_code	ret;	/* a return value		*/

	int 	num_in_queue,	/* number found in queue	*/
		rp;		/* running pointer		*/

	QUEUE_Element	*qep;	/* queue element pointer	*/

	static QUEUE_Element
			*save_ptr = (QUEUE_Element *) NULL;

/*
 * Gain access to the que so no one can change it...
 */
	if ( (ret = SNQ_Wait ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_DumpQueue: Semaphore failure\n");
#endif	
		return (ret);
	}

	if (save_ptr != (QUEUE_Element *) NULL)
		free (save_ptr);

	if (QPTRS->SNQ_head == QPTRS->SNQ_tail){
		if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
			fprintf(stderr,"SNQ_DumpQueue: Semaphore failure\n");
#endif	
			return (ret);
		}
#ifdef DEBUG
		fprintf(stderr,"SNQ_DumpQueue: SNQ queue empty\n");
#endif	
		return (SNQ_QUEUEEMPTY);
	}
	num_in_queue = 0;
	rp = QPTRS->SNQ_head;
	while (rp != QPTRS->SNQ_tail){
		num_in_queue++;
		if (++rp >= QPTRS->SNQ_size)
			rp = 0;
	}
	if ( (qep = (QUEUE_Element *) malloc (num_in_queue*
			sizeof (QUEUE_Element))) == (QUEUE_Element *)NULL ){
		if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
			fprintf(stderr,"SNQ_DumpQueue: Semaphore failure\n");
#endif	
			return (ret);
		}
#ifdef DEBUG
		fprintf(stderr,"SNQ_DumpQueue: No available memory\n");
#endif	
		return (SNQ_NOMEMORY);
	}
	*table = qep;
	rp = QPTRS->SNQ_head;
	while (rp != QPTRS->SNQ_tail){
		*qep = SNQ [rp]; qep++;
		if (++rp >= QPTRS->SNQ_size)
			rp = 0;
	}
	*num_active = num_in_queue;
	save_ptr = *table;
	
/*
 * Give the Queue Back
 */
	if ( (ret = SNQ_Signal ()) != SNQ_NORMAL ) {
#ifdef DEBUG
		fprintf(stderr,"SNQ_DumpQueue: Semaphore failure\n");
#endif	
		return (ret);
	}
	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_DumpQueue");
	return SNQ_UNIMPLEMENTED;
#endif
}


/* SNQ_PrintQueue
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_PrintQueue ( void ) {

#ifdef IPC_AVAILABLE
	int 	i;
	
	printf ("******************* SIZE: %d\n",QPTRS->SNQ_size);
	for (i = 0; i < QPTRS->SNQ_size; i++){
		printf ("%3d: %s  %d", i, SNQ [i].Q_ImageName,
			SNQ [i].Q_AssociationID);
		if ((QPTRS->SNQ_head == i) && (QPTRS->SNQ_tail == i))
			printf (" <-- HEAD/TAIL\n");
		else if (QPTRS->SNQ_head == i)
			printf (" <-- HEAD\n");
		else if (QPTRS->SNQ_tail == i)
			printf (" <-- TAIL\n");
		else
			printf ("\n");
	}
	printf ("*******************\n");
	return(SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_PrintQueue");
	return SNQ_UNIMPLEMENTED;
#endif
}



/* SNQ_Wait
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_Wait ( void ){

#ifdef IPC_AVAILABLE
	struct sembuf op;

	op.sem_num = 0;
	op.sem_op = -1;
	op.sem_flg = 0;	/* this was uninitialized - ajh 19/11/93 */
	if ((semop (SEM_ID, &op, (int) 1)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_Wait: semop failed\n");
#endif	
		return (SNQ_SEMAPHOREFAIL);
	}
	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_Wait");
	return SNQ_UNIMPLEMENTED;
#endif
}

	
/* SNQ_Signal
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

SNQ_code SNQ_Signal ( void ){

#ifdef IPC_AVAILABLE

	sem_buf.val = 1; /* this is the global buffer 	*/

	if ((semctl (SEM_ID, 0, SETVAL, sem_buf)) == -1){
#ifdef DEBUG
		fprintf(stderr,"SNQ_Signal: semctl failed\n");
#endif	
		return (SNQ_SEMAPHOREFAIL);
	}
	return (SNQ_NORMAL);
#else
	fprintf(stderr, SNQ_UNIMPLEMENTED_TXT,"SNQ_Signal");
	return SNQ_UNIMPLEMENTED;
#endif
}
	
/* SNQ_MakeFilename
**
** Modification History
** 
** who		   when			why
** -------------------------------------------------------------------------
** deb		08-Jul-1992		Original Code
** smm		10-Jul-1992		Add conditional compilation
**					for machines w/o shared mem.
*/

char	*SNQ_MakeFilename ( int qid ){

#ifdef IPC_AVAILABLE

	static char	buf[256];
	char		*p,
			temp[10];

	if (qid > MAX_NAMED_QUEUES){
		return (NULL);
	
	}
	if ((p = getenv (QUEUE_DIRECTORY)) == NULL){
		/*** return (NULL); ***/
		p = "/tmp";
	}
	strcpy (buf,p);
	if (buf[strlen(buf)-1] != '/')
		strcat (buf,"/");

	strcat (buf,QUEUE_FILESUFFIX);
	sprintf (temp,"%d",qid);
	strcat (buf,temp);

	return (buf);
	
#else
	return NULL;
#endif
}
